package be.jeffcheasey88.peeratcode.parser.state;

import java.util.ArrayList;
import java.util.List;
import java.util.function.BiFunction;
import java.util.function.Function;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.TokenValidator;

public class BuilderStateTree<E, B> extends StateTree<B>{
	
	private BiFunction<E, Bag, B> builder;
	
	private List<StateTree<E>> childs;
	private StateTree<E> multiple;

	public BuilderStateTree(BiFunction<E, Bag, B> builder){
		super();
		this.builder = builder;
		
		this.childs = new ArrayList<>();
	}
	
	B build(TokenValidator validator, E element){
		B build = this.builder.apply(element, validator.getBag());
		validator.setBag(new Bag());
		
		BuilderStateTree<B, ?> builded = super.internalSeed(validator, build);
		if(builded != null) builded.build(validator, build);
		
		internalSeedNoChilds(validator, element);
		
		return build;
	}
	
	void internalSeedNoChilds(TokenValidator validator, E element){
		if(multiple != null){
			TokenValidator branch = validator.branch();
			if(multiple.checker == null){
				BuilderStateTree<E, ?> builded;
				while((builded = multiple.internalSeed(branch, element)) != null){
					validator.merge(branch);
					builded.build(validator, element);
				}
			}else{
				while(multiple.checker.apply(branch)){
					BuilderStateTree<E, ?> builded = multiple.internalSeed(branch, element);
					if(builded != null){
						validator.merge(branch);
						builded.build(validator, element);
					}
				}
			}
		}
		for(StateTree<E> child : this.childs){
			TokenValidator branch = validator.branch();
			if(child.checker == null){
				BuilderStateTree<E, ?> builded = child.internalSeed(branch, element);
				if(builded != null){
					validator.merge(branch);
					builded.build(validator, element);
					return;
				}
				continue;
			}
			if(child.checker.apply(branch)){
				BuilderStateTree<E, ?> builded = child.internalSeed(branch, element);
				if(builded != null){
					validator.merge(branch);
					builded.build(validator, element);
					return;
				}
			}
		}
	}
	
	public <T extends StateTree<E>> StateTree<E> thenNoChild(StateTree<E> child){
		childs.add(child);
		return child;
	}
	
	public <T extends StateTree<E>> StateTree<E> thenNoChild(Function<TokenValidator, Boolean> checker){
		StateTree<E> state = new StateTree<>();
		state.checker = checker;
		this.childs.add(state);
		return state;
	}
	
	public <T extends StateTree<E>> StateTree<E> multiple(StateTree<E> child){
		this.multiple = child;
		return child;
	}
}
